/* Coloring labelled grid using NC colors - greedy approach. */
/* Labels are int's in [0..number of cells].		     */
/* Successfully colored regions are colored with [1..NC].    */
/* Other regions are colored with 0.		 	     */
/* Requires NC <= 32  (number of bits for int)		     */
/*    cc - region labels				     */
/*   *cy - set to number of cycles for coloring		     */
/*    result - coloring					     */
/* Uses "randomized" region labels.  Grid size must be a     */
/* power of 2.						     */
/* File:  grrclr.cs					     */

#include <stdio.h>
#include <stdlib.h>
#include <cscomm.h>

/* Internal coloring representation (sets of colors):	     */
/*							     */
/* xl:     0      = not yet colored			     */
/*	  2^(i-1) = color i				     */
/*	  -1      = uncolorable				     */
/*							     */
/* xw, xws:  sets of colors = logior(xl's), but not -1's     */

/* Included for testing */

void dcc(int:current *cx, int i, int j, int r, int c);

int:current grcolor_growing(int:current cc_base, int NC, int *cy, 
			  int:current root_clr)
 { int:current cc, sz, cl, sw[4], cw[4], mxm, mxr, xl=0, oldxl, xw, xws;
   CMC_sendaddr_t:current root;
   int W, more, u, twou, Nx, cyc;
  /* To loop over directions (N, E, S, W) */
   static int ax[4] = {1, 0, 1, 0}, di[4] = {-1, 1, 1, -1};
   printf ("\n**** I am in GRCOLOR now *****%d\n", [0][0]xl);

/*   printf("Using %d colors.\n", NC);	 Testing */
/* Randomize labels */
   with (current) everywhere
     cc = 128*pcoord(0) + pcoord(1);
   W  = dimof(current, 1);
   Nx = W*dimof(current, 0);	/* Grid size - POWER OF 2! */
   cl = prand();		/* Lots of random numbers */
   Nx = (Nx - 1) & (^= cl);	/* Random, 0 <= Nx < Grid size (XOR) */
   cc ^= Nx;			/* Flip random bits in labels */
   printf ("\n **** I did the random thingy ****%d\n", [0][0]xl);
  /* Locate region "roots" */
   root = make_send_address(current, cc%W, cc/W);
   printf ("\n **** I found all the cells roots *****%d\n", [0][0]xl);
  /* Calculate region sizes */
   sz = 0;
   send(&sz, root, (int:current)1, CMC_combiner_add, CMC_no_field);
   sz = get(root, &sz, CMC_collisions);
   printf ("\n**** Finding out how big the regions are ***%d\n", [0][0]xl);
/*   puts("Sizes:");	 Testing */
/*   dcc(&sz, 0, 0, 5, 12);      */
  /* Get neighbor labels, sizes */
  for ( u = 0; u < 4; u++ )
   { cw[u] = from_grid_dim(&cc, cc, ax[u], di[u]);
     sw[u] = from_grid_dim(&sz, sz, ax[u], di[u]);
   }
   printf ("\n***** Found all my neighbors stuff *****%d\n", [0][0]xl);
  /* Initialize coloring */
   xl = 0;	/* Region color (bitvector) */
   more = 1;	/* To start loop */
   Nx = 0;      /* Limit on no. of colors used so far */
   cyc = 0;	/* Cycle counter */
   printf ("\n**** Color was Initialized*****%d\n", [0][0]xl);
   where (((pcoord(0) * W) + pcoord(1)) == cc) 
     xl = 1 << (root_clr - 1);
   printf ("\n**** Roots were given final color *****%d\n", [0][0]xl);
   while ( more )	/* Try to color more regions */
    { printf ("\n**** Finding the largest one ****%d\n", [0][0]xl);
     oldxl = xl;	/* To detect changes */
     printf ("\n**** Finding the largest one ****%d\n", [0][0]xl);
     cyc++;
     printf ("\n**** Finding the largest one ****%d\n", [0][0]xl);
    /* Identify local maxima, get neighbor colors */
     printf ("\n**** Finding the largest one ****%d\n",[0][0]xl);
     mxr = -1;			/* Is region a local maximum? */
     printf ("\n**** Finding the largest one ****%d\n", [0][0]xl);
     mxm = (xl == 0);		/* Is cell a local maximum */
     printf ("\n**** Finding the largest one ****%d\n",[0][0]xl);
     xws = 0;			/* Set of neighbor colors */
     printf ("\n**** Finding the largest one ****\n");
     for ( u = 0; u < 4; u++ )	/* Check neighbors */
      { xw = from_grid_dim(&xl, xl, ax[u], di[u]);
	where ( xw != -1 ) xws |= xw;
	where ( xw == 0 )
	  mxm &=  (sw[u] < sz) | ((sw[u] == sz) & (cw[u] <= cc));
      }
     printf ("\n**** Found the largest one ****\n");
    /* puts("xws");	 Testing */
    /* dcc(&xws, 0, 0, 5, 12); */
     where ( xl == 0 )		/* Identify local max regions */
      { where ( !mxm )
	  send(&mxr, root, mxm, CMC_combiner_logand, CMC_no_field);
	else mxm = get(root, &mxr, CMC_collisions);
      }
     printf ("\n**** I got it now!, the largest one ****\n");
    /* Now mxm marks cells of locally max uncolored regions */
     mxr = 0;			/* For set of adjoining colors */
     if ( Nx < NC ) Nx++;	/* Allow one more color */
    /* printf("Nx = %d, mxm:\n", Nx);	 Testing */
    /* dcc(&mxm, 0, 0, 5, 12); */
     printf ("\n**** Trying to find the color ****\n");
     where ( mxm )		/* Determine color if possible */
     { send(&mxr, root, xws, CMC_combiner_logior, CMC_no_field);
       xws = get(root, &mxr, CMC_collisions);
       for ( u = 0, twou = 1;  u < Nx;  u++, twou <<= 1 )
       { where ( (xl == 0) & ((xws&1) == 0) ) xl = twou;
	 xws >>= 1;
       }
       where ( xl == 0 ) xl = -1;	/* Uncolorable */
     printf ("\n**** still looking ****\n");
     }
     printf ("\n**** Found the color ***\n");
    /* puts("xl:");	 Testing */
    /* dcc(&xl, 0, 0, 5, 12); */
     more = (|= (oldxl != xl));		/* Any changes? */
   } /* end while ( more )
  /* Calculate color numbers */
   xw = xl;  cl = 0;
   printf ("\n**** I actually am done ****\n");
   for ( u = 1, twou = 1;  u <= Nx;  u++, twou <<= 1 )
     where ( xw == twou ) cl = u;
   where ( xl == -1 ) cl = 0;
   *cy = cyc;		/* Report cycles */
   return (cl);		/* Region color (integer) */
 }

int:current grcolor( int:current cc, int NC, int *cy )
 { int:current sz, cl, sw[4], cw[4], mxm, mxr, xl, oldxl, xw, xws;
   CMC_sendaddr_t:current root;
   int W, more, u, twou, Nx, cyc;
  /* To loop over directions (N, E, S, W) */
   static int ax[4] = {1, 0, 1, 0}, di[4] = {-1, 1, 1, -1};

/*   printf("Using %d colors.\n", NC);	 Testing */
/* Randomize labels */
   W  = dimof(current, 1);
   Nx = W*dimof(current, 0);	/* Grid size - POWER OF 2! */
   cl = prand();		/* Lots of random numbers */
   Nx = (Nx - 1) & (^= cl);	/* Random, 0 <= Nx < Grid size (XOR) */
   cc ^= Nx;			/* Flip random bits in labels */
/* Locate region "roots" */
   root = make_send_address(current, cc%W, cc/W);
  /* Calculate region sizes */
   sz = 0;
   send(&sz, root, (int:current)1, CMC_combiner_add, CMC_no_field);
   sz = get(root, &sz, CMC_collisions);
/*   puts("Sizes:");	 Testing */
/*   dcc(&sz, 0, 0, 5, 12);      */
  /* Get neighbor labels, sizes */
  for ( u = 0; u < 4; u++ )
   { cw[u] = from_grid_dim(&cc, cc, ax[u], di[u]);
     sw[u] = from_grid_dim(&sz, sz, ax[u], di[u]);
   }
  /* Initialize coloring */
   xl = 0;	/* Region color (bitvector) */
   more = 1;	/* To start loop */
   Nx = 0;      /* Limit on no. of colors used so far */
   cyc = 0;	/* Cycle counter */
   while ( more )	/* Try to color more regions */
   { oldxl = xl;	/* To detect changes */
     cyc++;
    /* Identify local maxima, get neighbor colors */
     mxr = -1;			/* Is region a local maximum? */
     mxm = (xl == 0);		/* Is cell a local maximum */
     xws = 0;			/* Set of neighbor colors */
     for ( u = 0; u < 4; u++ )	/* Check neighbors */
      { xw = from_grid_dim(&xl, xl, ax[u], di[u]);
	where ( xw != -1 ) xws |= xw;
	where ( xw == 0 )
	  mxm &=  (sw[u] < sz) | ((sw[u] == sz) & (cw[u] <= cc));
      }
    /* puts("xws");	 Testing */
    /* dcc(&xws, 0, 0, 5, 12); */
     where ( xl == 0 )		/* Identify local max regions */
      { where ( !mxm )
	  send(&mxr, root, mxm, CMC_combiner_logand, CMC_no_field);
	else mxm = get(root, &mxr, CMC_collisions);
      }
    /* Now mxm marks cells of locally max uncolored regions */
     mxr = 0;			/* For set of adjoining colors */
     if ( Nx < NC ) Nx++;	/* Allow one more color */
    /* printf("Nx = %d, mxm:\n", Nx);	 Testing */
    /* dcc(&mxm, 0, 0, 5, 12); */
     where ( mxm )		/* Determine color if possible */
     { send(&mxr, root, xws, CMC_combiner_logior, CMC_no_field);
       xws = get(root, &mxr, CMC_collisions);
       for ( u = 0, twou = 1;  u < Nx;  u++, twou <<= 1 )
       { where ( (xl == 0) & ((xws&1) == 0) ) xl = twou;
	 xws >>= 1;
       }
       where ( xl == 0 ) xl = -1;	/* Uncolorable */
     }
    /* puts("xl:");	 Testing */
    /* dcc(&xl, 0, 0, 5, 12); */
     more = (|= (oldxl != xl));		/* Any changes? */
   } /* end while ( more )
  /* Calculate color numbers */
   xw = xl;  cl = 0;
   for ( u = 1, twou = 1;  u <= Nx;  u++, twou <<= 1 )
     where ( xw == twou ) cl = u;
   where ( xl == -1 ) cl = 0;
   *cy = cyc;		/* Report cycles */
   return (cl);		/* Region color (integer) */
 }

int:current grcolor_moving( int:current cc, int NC, int *cy )
 { int:current sz, cl, sw[4], cw[4], mxm, mxr, xl, oldxl, xw, xws;
   CMC_sendaddr_t:current root;
   int W, more, u, twou, Nx, cyc;
  /* To loop over directions (N, E, S, W) */
   static int ax[4] = {1, 0, 1, 0}, di[4] = {-1, 1, 1, -1};

/*   printf("Using %d colors.\n", NC);	 Testing */
/* Randomize labels */
   W  = dimof(current, 1);
   Nx = W*dimof(current, 0);	/* Grid size - POWER OF 2! */
   cl = prand();		/* Lots of random numbers */
   Nx = (Nx - 1) & (^= cl);	/* Random, 0 <= Nx < Grid size (XOR) */
   cc ^= Nx;			/* Flip random bits in labels */
/* Locate region "roots" */
   root = make_send_address(current, cc%W, cc/W);
  /* Calculate region sizes */
   sz = 0;
   send(&sz, root, (int:current)1, CMC_combiner_add, CMC_no_field);
   sz = get(root, &sz, CMC_collisions);
/*   puts("Sizes:");	 Testing */
/*   dcc(&sz, 0, 0, 5, 12);      */
  /* Get neighbor labels, sizes */
  for ( u = 0; u < 4; u++ )
   { cw[u] = from_grid_dim(&cc, cc, ax[u], di[u]);
     sw[u] = from_grid_dim(&sz, sz, ax[u], di[u]);
   }
  /* Initialize coloring */
   xl = 0;	/* Region color (bitvector) */
   more = 1;	/* To start loop */
   Nx = 0;      /* Limit on no. of colors used so far */
   cyc = 0;	/* Cycle counter */
   while ( more )	/* Try to color more regions */
   { oldxl = xl;	/* To detect changes */
     cyc++;
    /* Identify local maxima, get neighbor colors */
     mxr = -1;			/* Is region a local maximum? */
     mxm = (xl == 0);		/* Is cell a local maximum */
     xws = 0;			/* Set of neighbor colors */
     for ( u = 0; u < 4; u++ )	/* Check neighbors */
      { xw = from_grid_dim(&xl, xl, ax[u], di[u]);
	where ( xw != -1 ) xws |= xw;
	where ( xw == 0 )
	  mxm &=  (sw[u] < sz) | ((sw[u] == sz) & (cw[u] <= cc));
      }
    /* puts("xws");	 Testing */
    /* dcc(&xws, 0, 0, 5, 12); */
     where ( xl == 0 )		/* Identify local max regions */
      { where ( !mxm )
	  send(&mxr, root, mxm, CMC_combiner_logand, CMC_no_field);
	else mxm = get(root, &mxr, CMC_collisions);
      }
    /* Now mxm marks cells of locally max uncolored regions */
     mxr = 0;			/* For set of adjoining colors */
     if ( Nx < NC ) Nx++;	/* Allow one more color */
    /* printf("Nx = %d, mxm:\n", Nx);	 Testing */
    /* dcc(&mxm, 0, 0, 5, 12); */
     where ( mxm )		/* Determine color if possible */
     { send(&mxr, root, xws, CMC_combiner_logior, CMC_no_field);
       xws = get(root, &mxr, CMC_collisions);
       for ( u = 0, twou = 1;  u < Nx;  u++, twou <<= 1 )
       { where ( (xl == 0) & ((xws&1) == 0) ) xl = twou;
	 xws >>= 1;
       }
       where ( xl == 0 ) xl = -1;	/* Uncolorable */
     }
    /* puts("xl:");	 Testing */
    /* dcc(&xl, 0, 0, 5, 12); */
     more = (|= (oldxl != xl));		/* Any changes? */
     if (more)
       { xw = xl; cl = 0;
	 where (xl == 0)
	   cl = -2;
	 else
	   { for ( u = 1, twou = 1;  u <= Nx;  u++, twou <<= 1 )
	       where ( xw == twou ) cl = u;
	     where ( xl == -1 ) cl = 0;
	   }
	 Plotgclr(cl);
       }
   } /* end while ( more )
  /* Calculate color numbers */
   xw = xl;  cl = 0;
   for ( u = 1, twou = 1;  u <= Nx;  u++, twou <<= 1 )
     where ( xw == twou ) cl = u;
   where ( xl == -1 ) cl = 0;
   *cy = cyc;		/* Report cycles */
   return (cl);		/* Region color (integer) */
 }

